"""Helper functions/classes, for DumpDie example views.

Exists here, so that the project only imports these on example view load.
Helps prevent package errors relating to example view load from propagating
to general package usage.
"""

# System Imports.
import copy
import os
from datetime import datetime, timedelta
from decimal import Decimal
from enum import Enum
from pathlib import Path, PosixPath, PurePath, WindowsPath
from types import ModuleType

# Third-Party Imports.
from django.core.files import File
from django.db import models
from django.forms import ModelForm
from django.http import QueryDict
from django.shortcuts import render
from django.template.response import TemplateResponse
from django.utils import timezone

# Internal Imports.
from django_dump_die.constants import PYTZ_PRESENT, ZONEINFO_PRESENT

# Imports that may not be accessible, depending on local python environment setup.
if PYTZ_PRESENT:
    import pytz
if ZONEINFO_PRESENT:
    from zoneinfo import ZoneInfo


BASE_DIR = os.path.dirname(os.path.abspath(__file__))


# region Helper Functions/Classes

def sample_func():
    """Sample doc string"""
    return 42


def sample_func_param(param1, *args, some_kwarg=None, **kwargs):
    """Sample param doc string.

    :param param1: Doc for param1.
    """
    return 'MyReturnValue with param1 as: {0}'.format(param1)


class EmptyClass:
    """Empty sample class"""
    pass


class SimpleClass:
    """Spam sample class."""

    SAMPLE_CLASS_CONST = 'Sample Class Constant Content'

    sample_class_var = 23

    def __init__(self):
        self._sample_private_module = ModuleType('django.html')
        self._sample_private_bytes = b'sample bytes'
        self._sample_private_date = datetime.now().date()
        self._sample_private_datetime = datetime.now()
        self._sample_private_time = datetime.now().time()
        self._sample_private_int = 42
        self._sample_private_float = 42.42
        self._sample_private_decimal = Decimal(42.42)
        self._sample_private_string = 'Sample String Content'
        self._sample_private_none = None
        self._sample_private_bool = True

        self._sample_private_set = {'A', 'B', 'C'}
        self._sample_private_tuple = ('A', 12, True)
        self._sample_private_list = ['A', 12, True]
        self._sample_private_dict = {
            'first': 'A',
            'second': 12,
            'third': True,
        }

        self.sample_public_module = ModuleType('django.html')
        self.sample_public_bytes = b'sample bytes'
        self.sample_public_date = datetime.now().date()
        self.sample_public_datetime = datetime.now()
        self.sample_public_time = datetime.now().time()
        self.sample_public_int = 42
        self.sample_public_float = 42.42
        self.sample_public_decimal = Decimal(42.42)
        self.sample_public_string = 'Sample String Content'
        self.sample_public_none = None
        self.sample_public_bool = True

        self.sample_public_set = {'A', 'B', 'C'}
        self.sample_public_tuple = ('A', 12, True)
        self.sample_public_list = ['A', 12, True]
        self.sample_public_dict = {
            'first': 'A',
            'second': 12,
            'third': True,
        }

    def sample_class_func(self):
        """Sample class func doc string"""
        return 'Sample class result'

    def sample_class_param_func(self, arg1, *args, **kwargs):
        """Sample class param func doc string"""
        return arg1


class ComplexClass:
    """Complex class with nested instances"""

    def __init__(self):
        self._sample_private_simple_class = SimpleClass()
        self.sample_public_simple_class = SimpleClass()
        self.duplicate_sample_public_simple_class = self.sample_public_simple_class


class SampleEnum(Enum):
    """Sample Enum"""
    RED = 1
    BLUE = 2


class SampleRelation(models.Model):
    """Sample Class for foreign relation"""
    relate_name = models.TextField()


class SampleManyRelation(models.Model):
    """Sample Class for many to many relation"""
    relate_name = models.TextField()


class SampleOneRelation(models.Model):
    """Sample Class for one to one relation"""
    relate_name = models.TextField()


class SampleDjangoModel(models.Model):
    """Sample Django Model with all field types.

    To be backwards-compatible with all possible Django versions, we check for attr errors
    and skip field if error is present.

    Note: We add this logic to all fields, to account for both past-Django versions not
    having some fields, as well as hypothetical future versions of Django where one or
    more of these fields may become depreciated.
    """

    try:
        sample_big_int = models.BigIntegerField()
    except AttributeError:
        pass

    try:
        sample_binary = models.BinaryField()
    except AttributeError:
        pass

    try:
        sample_bool = models.BooleanField()
    except AttributeError:
        pass

    try:
        sample_char = models.CharField(max_length=200)
    except AttributeError:
        pass

    try:
        sample_date = models.DateField()
    except AttributeError:
        pass

    try:
        sample_datetime = models.DateTimeField()
    except AttributeError:
        pass

    try:
        sample_decimal = models.DecimalField(decimal_places=8, max_digits=16)
    except AttributeError:
        pass

    try:
        sample_duration = models.DurationField()
    except AttributeError:
        pass

    try:
        sample_email = models.EmailField()
    except AttributeError:
        pass

    try:
        sample_file = models.FileField(upload_to='uploads')
    except AttributeError:
        pass

    try:
        sample_file_path = models.FilePathField(path='/')
    except AttributeError:
        pass

    try:
        sample_float = models.FloatField()
    except AttributeError:
        pass

    try:
        sample_ip = models.GenericIPAddressField()
    except AttributeError:
        pass

    try:
        sample_image = models.ImageField(upload_to='uploads')
    except AttributeError:
        pass

    try:
        sample_int = models.IntegerField()
    except AttributeError:
        pass

    try:
        sample_json = models.JSONField()
    except AttributeError:
        pass

    try:
        sample_pos_bint = models.PositiveBigIntegerField()
    except AttributeError:
        pass

    try:
        sample_pos_int = models.PositiveIntegerField()
    except AttributeError:
        pass

    try:
        sample_pos_sint = models.PositiveSmallIntegerField()
    except AttributeError:
        pass

    try:
        sample_sint = models.SmallIntegerField()
    except AttributeError:
        pass

    try:
        sample_slug = models.SlugField()
    except AttributeError:
        pass

    try:
        sample_text = models.TextField()
    except AttributeError:
        pass

    try:
        sample_time = models.TimeField()
    except AttributeError:
        pass

    try:
        sample_url = models.URLField()
    except AttributeError:
        pass

    try:
        sample_uuid = models.UUIDField()
    except AttributeError:
        pass

    try:
        sample_foreign = models.ForeignKey(SampleRelation, on_delete=models.CASCADE, related_name='sample_foreign')
    except AttributeError:
        pass

    try:
        sample_many = models.ManyToManyField(SampleManyRelation, 'sample_many')
    except AttributeError:
        pass

    try:
        sample_one = models.OneToOneField(SampleOneRelation, on_delete=models.CASCADE, related_name='sample_one')
    except AttributeError:
        pass


class SampleModelForm(ModelForm):
    """Sample Model Form.

    To be backwards-compatible with all possible Django versions, we check for attr existence
    and include field if respective attr is present.

    Note: We add this logic to all fields, to account for both past-Django versions not
    having some fields, as well as hypothetical future versions of Django where one or
    more of these fields may become depreciated.
    """
    class Meta:
        """Meta info"""
        model = SampleDjangoModel
        fields = []

        if hasattr(SampleDjangoModel, 'sample_big_int'):
            fields += ['sample_big_int']
        if hasattr(SampleDjangoModel, 'sample_bool'):
            fields += ['sample_bool']
        if hasattr(SampleDjangoModel, 'sample_char'):
            fields += ['sample_char']
        if hasattr(SampleDjangoModel, 'sample_date'):
            fields += ['sample_date']
        if hasattr(SampleDjangoModel, 'sample_datetime'):
            fields += ['sample_datetime']
        if hasattr(SampleDjangoModel, 'sample_decimal'):
            fields += ['sample_decimal']
        if hasattr(SampleDjangoModel, 'sample_duration'):
            fields += ['sample_duration']
        if hasattr(SampleDjangoModel, 'sample_email'):
            fields += ['sample_email']
        if hasattr(SampleDjangoModel, 'sample_file'):
            fields += ['sample_file']
        if hasattr(SampleDjangoModel, 'sample_file_path'):
            fields += ['sample_file_path']
        if hasattr(SampleDjangoModel, 'sample_float'):
            fields += ['sample_float']
        if hasattr(SampleDjangoModel, 'sample_ip'):
            fields += ['sample_ip']
        if hasattr(SampleDjangoModel, 'sample_image'):
            fields += ['sample_image']
        if hasattr(SampleDjangoModel, 'sample_int'):
            fields += ['sample_int']
        if hasattr(SampleDjangoModel, 'sample_json'):
            fields += ['sample_json']
        if hasattr(SampleDjangoModel, 'sample_pos_bint'):
            fields += ['sample_pos_bint']
        if hasattr(SampleDjangoModel, 'sample_pos_int'):
            fields += ['sample_pos_int']
        if hasattr(SampleDjangoModel, 'sample_pos_sint'):
            fields += ['sample_pos_sint']
        if hasattr(SampleDjangoModel, 'sample_slug'):
            fields += ['sample_slug']
        if hasattr(SampleDjangoModel, 'sample_sint'):
            fields += ['sample_sint']
        if hasattr(SampleDjangoModel, 'sample_text'):
            fields += ['sample_text']
        if hasattr(SampleDjangoModel, 'sample_time'):
            fields += ['sample_time']
        if hasattr(SampleDjangoModel, 'sample_url'):
            fields += ['sample_url']
        if hasattr(SampleDjangoModel, 'sample_uuid'):
            fields += ['sample_uuid']
        if hasattr(SampleDjangoModel, 'sample_foreign'):
            fields += ['sample_foreign']
        if hasattr(SampleDjangoModel, 'sample_many'):
            fields += ['sample_many']
        if hasattr(SampleDjangoModel, 'sample_one'):
            fields += ['sample_one']

# endregion Helper Functions/Classes


# region DD Display Functions

def dump_simple_types():
    """Dump Simple Types"""
    # Generate variables to dump.
    SAMPLE_CONST = 'Sample Constant Content'
    sample_module = ModuleType('django.html')
    sample_bytes = b'sample bytes'
    sample_int = 42
    sample_float = 42.42
    sample_decimal = Decimal(42.42)
    sample_string = 'Sample String Content'
    sample_none = None
    sample_bool = True

    # Call dump on all generated variables.
    dump('')
    dump(SAMPLE_CONST)
    dump(sample_module)
    dump(sample_bytes)
    dump(sample_int)
    dump(sample_float)
    dump(sample_decimal)
    dump(sample_string)
    dump(sample_none)
    dump(sample_bool)


def dump_intermediate_types():
    """Dump Intermediate Types"""
    # Generate variables to dump.
    sample_bytes_array = bytearray([8, 9, 10, 11])
    sample_complex = 3 - 1j

    # Call dump on all generated variables.
    dump('')
    dump('Python type examples:')
    dump(sample_bytes_array)
    dump(sample_complex)

    dump_datetime_types()

    # Generate variables to dump.
    sample_pure_path = PurePath(Path.cwd())
    try:
        sample_posix_path = PosixPath(Path.cwd())
    except NotImplementedError:
        sample_posix_path = None
    try:
        sample_windows_path = WindowsPath(Path.cwd())
    except NotImplementedError:
        sample_windows_path = None

    # Call dump on all generated variables.
    dump('')
    dump('Python pathlib examples:')
    dump(sample_pure_path)
    if sample_posix_path:
        dump(sample_posix_path)
    if sample_windows_path:
        dump(sample_windows_path)


class dump_complex_types():
    """Dump Complex Types"""

    sample_multilevel_set = {
        (
            'A',
            12,
            True,
        ),
        (
            'B',
            24,
            False,
        ),
    }

    sample_multilevel_tuple = (
        {
            'first': 'A',
            'second': 12,
            'third': True,
        },
        {
            'fourth': 'B',
            'fifth': 24,
            'sixth': False,
        },
    )

    sample_multilevel_list = [
        {
            'first': 'A',
            'second': 12,
            'third': True,
        },
        {
            'fourth': 'B',
            'fifth': 24,
            'sixth': False,
        },
    ]

    sample_multilevel_dict = {
        'initial': {
            'first': 'A',
            'second': 12,
            'third': True,
        },
        'secondary': {
            'fourth': 'B',
            'fifth': 24,
            'sixth': False,
        },
    }

    def dump_set(self):
        # Generate variables to dump.
        sample_set = {'A', 'B', 'C'}

        # Dump variable.
        dump(sample_set)

    def dump_frozen_set(self):
        # Generate variables to dump.
        sample_frozen_set = frozenset({'D', 'E', 'F'})

        # Dump variable.
        dump(sample_frozen_set)

    def dump_tuple(self):
        # Generate variables to dump.
        sample_tuple = ('A', 12, True)

        # Dump variable.
        dump(sample_tuple)

    def dump_list(self):
        # Generate variables to dump.
        sample_list = ['A', 12, True]

        # Dump variable.
        dump(sample_list)

    def dump_dict(self):
        # Generate variables to dump.
        sample_dict = {
            'first': 'A',
            'second': 12,
            'third': True,
        }

        # Dump variable.
        dump(sample_dict)

    def dump_querydict(self):
        # Generate variables to dump.
        sample_querydict = QueryDict('first="A"&second=12&third=True&first="B"&first="C"')

        # Dump variable.
        dump(sample_querydict)

    def dump_enum(self):
        # Dump variable.
        dump(SampleEnum)

    def dump_memory_view(self):
        # Generate variables to dump.
        sample_memory_view = memoryview(bytearray([8, 9, 10, 11]))

        # Dump variable.
        dump(sample_memory_view)

    def dump_multilevel_set(self):
        # Dump variable.
        dump(self.sample_multilevel_set)

    def dump_multilevel_tuple(self):
        # Dump variable.
        dump(self.sample_multilevel_tuple)

    def dump_multilevel_list(self):
        dump(self.sample_multilevel_list)

    def dump_multilevel_dict(self):
        # Dump variable.
        dump(self.sample_multilevel_dict)

    def dump_list_subitem(self):
        # Dump variable.
        dump(self.sample_multilevel_list[0])

    def dump_tuple_subitem(self):
        # Dump variable.
        dump(self.sample_multilevel_tuple[0])

    def dump_tuple_subitem_function(self):
        # Dump variable.
        dump(self.sample_multilevel_tuple[0].items)

    def dump_dict_subitem(self):
        # Dump variable.
        dump(self.sample_multilevel_dict['initial'])

    def dump_enum_subitem(self):
        # Dump variable.
        dump(SampleEnum.RED)
        dump(SampleEnum.BLUE)

    def dump_all_objects(self):
        # Call dump on all generated variables.
        dump('')
        dump('Minimal object examples:')
        self.dump_set()
        self.dump_frozen_set()
        self.dump_tuple()
        self.dump_list()
        self.dump_dict()
        self.dump_querydict()
        self.dump_memory_view()
        self.dump_enum()

        dump('')
        dump('Nested object examples:')
        self.dump_multilevel_set()
        self.dump_multilevel_tuple()
        self.dump_multilevel_list()
        self.dump_multilevel_dict()

        dump('')
        dump('Examples of pulling items/indexes/subsets from above objects:')
        self.dump_list_subitem()
        self.dump_tuple_subitem()
        self.dump_tuple_subitem_function()
        self.dump_dict_subitem()
        self.dump_enum_subitem()


def dump_function_types():
    """Dump Function Types"""
    # Generate variables to dump.
    # None for this view.

    # Call dump on all generated variables.
    dump('')
    dump('Function examples:')
    # Minimal function with no args.
    dump(sample_func)
    # Function with args & kwargs.
    dump(sample_func_param)

    dump('')
    dump('Function call examples:')
    # Calling above "minimal function with no args".
    dump(sample_func())
    # Calling function with one arg.
    dump(sample_func_param(32))
    # Calling function with both args & kwargs.
    dump(sample_func_param('test_param', some_kwarg=True))
    # Calling function with multiple args.
    dump(sample_func_param('test_param', 'extra_arg_1', 2, True))


def dump_class_types():
    """Dump Class Types"""
    # Generate variables to dump.
    sample_empty_class = EmptyClass()
    sample_simple_class = SimpleClass()
    sample_complex_class = ComplexClass()

    # Call dump on all generated variables.
    dump('')
    dump('Class object examples:')
    dump(EmptyClass)
    dump(SimpleClass)
    dump(ComplexClass)

    dump('')
    dump('Class instance examples:')
    dump(sample_empty_class)
    dump(sample_simple_class)
    dump(sample_complex_class)
    dump(sample_complex_class)

    dump('')
    dump('Examples of pulling nested items (classes/functions/data/etc) from above classes.')
    dump(sample_complex_class._sample_private_simple_class)
    dump(sample_complex_class.sample_public_simple_class)
    dump(sample_complex_class.sample_public_simple_class.sample_public_dict)
    dump(sample_complex_class.sample_public_simple_class.sample_public_dict['first'])
    dump(sample_simple_class.sample_class_func)
    dump(sample_simple_class.sample_class_param_func)
    dump(sample_simple_class.sample_class_func())
    dump(sample_simple_class.sample_class_param_func('test'))


def dump_numeric_types():
    """Dump Numeric Types"""
    # Generate variables to dump.
    sample_int = 42
    sample_float = 42.42
    sample_decimal = Decimal(42.42)
    sample_complex = 3 - 1j

    # Call dump on all generated variables.
    dump('')
    dump('Numeric examples:')
    dump(sample_int)
    dump(sample_float)
    dump(sample_decimal)
    dump(sample_complex)


def dump_datetime_types():
    """Dump Datetime Types"""
    # Generate variables to dump.
    sample_dt_date = datetime.now().date()
    sample_tz_date = timezone.now().date()
    sample_dt_datetime = datetime.now()
    sample_tz_datetime = timezone.now()
    sample_dt_time = datetime.now().time()
    sample_tz_time = timezone.now().time()
    sample_dt_timedelta = timedelta(days=1)
    sample_tz_timedelta = timezone.timedelta(days=2)
    if PYTZ_PRESENT:
        sample_pytz_timezone = pytz.timezone('UTC')
    if ZONEINFO_PRESENT:
        sample_zoneinfo_timezone = ZoneInfo('UTC')

    # Call dump on all generated variables.
    dump('')
    dump('Date/Time examples:')
    dump(sample_dt_date)
    dump(sample_tz_date)
    dump(sample_dt_datetime)
    dump(sample_tz_datetime)
    dump(sample_dt_time)
    dump(sample_tz_time)
    dump(sample_dt_timedelta)
    dump(sample_tz_timedelta)
    if PYTZ_PRESENT:
        dump(sample_pytz_timezone)
    if ZONEINFO_PRESENT:
        dump(sample_zoneinfo_timezone)


def dump_model_types():
    """Dump Model Types"""
    # Generate variables to dump.
    sample_django_model_empty = SampleDjangoModel(id=1)  # Must provide id so many to many field can be dumped.
    sample_model_form = SampleModelForm()

    # Call dump on all generated variables.
    dump('')
    dump('Model and ModelForm class examples:')
    # TODO: See if there is a better way to "fix" this.
    # Need to remove the Many-To-Many relationship when dumping the class
    # definition (non-instance) because python's get_members can't handle a
    # m2m when the "object" does not have an id. Which it can't due to it being
    # a class definition and not an instance. Problem is solved when an instance
    # is created. See below dump of instance.
    SampleDjangoModelNoManyRelation = copy.deepcopy(SampleDjangoModel)
    delattr(SampleDjangoModelNoManyRelation, 'sample_many')
    dump(SampleDjangoModelNoManyRelation)

    dump(SampleModelForm)

    dump('')
    dump('Model and ModelForm instance examples:')
    dump(sample_django_model_empty)
    dump(sample_model_form)

    # Populate model with values and re-check.
    dump('')
    dump('Example of model with all values populated/updated from above.')
    path = os.path.join(BASE_DIR, '../../media/uploads/test_img.png')
    with open(path, 'rb') as local_file:
        django_file = File(local_file, name=os.path.basename(local_file.name))

        sample_django_model_populated = SampleDjangoModel(
            pk=1,
            sample_big_int=1,
            sample_binary=b'',
            sample_bool=True,
            sample_char='A',
            sample_date=datetime.now().date(),
            sample_datetime=datetime.now(),
            sample_decimal=Decimal(42.42),
            sample_duration=None,
            sample_email='someone@example.com',
            sample_file=django_file,
            sample_file_path='../media',
            sample_float=42.42,
            sample_image=django_file,
            sample_int=5,
            sample_ip='127.0.0.1',
            # sample_json='{"key": "my_val"}',
            # sample_pos_bint=345,
            sample_pos_int=34,
            sample_pos_sint=3,
            sample_sint=3,
            sample_slug='foobar',
            sample_text='All my text',
            sample_time=datetime.now().time(),
            sample_url='https://github.com',
            sample_uuid='asdfasdf',
        )

        dump(sample_django_model_populated)


def dump_iterable_group_types():
    """Dump Iterable Group Types"""
    dump_complex_types()


def dump_syspath_types():
    """Dump System Path Types"""
    # Generate variables to dump.
    os_path = os.path.abspath(os.getcwd())
    pure_path = PurePath(Path.cwd())
    try:
        posix_path = PosixPath(Path.cwd())
    except NotImplementedError:
        posix_path = None
    try:
        windows_path = WindowsPath(Path.cwd())
    except NotImplementedError:
        windows_path = None

    # Call dump on all generated variables.
    dump('')
    dump('Python os.path examples:')
    dump(os.path)
    dump(os_path)

    dump('')
    dump('Python pathlib examples:')
    dump(pure_path)
    if posix_path:
        dump(posix_path)
    if windows_path:
        dump(windows_path)


def dump_django_request_response_cycle_types(request):
    """Dump Django request-response cycle Types."""

    # Generate variables to dump.
    sample_query_dict = QueryDict('one_val=test&two_vals=one&two_vals=2')
    sample_query_dict = copy.deepcopy(sample_query_dict)
    sample_query_dict.appendlist('example_field_list', 'username')
    sample_query_dict.appendlist('example_field_list', 'first_name')
    sample_query_dict.appendlist('example_field_list', 'last_name')
    sample_query_dict.appendlist('example_types', None)
    sample_query_dict.appendlist('example_types', True)
    sample_query_dict.appendlist('example_types', 5)
    sample_query_dict.appendlist('example_types', 3.0)

    sample_request = request
    sample_http_response = render(request, 'django_dump_die/sample.html', {})
    sample_template_response = TemplateResponse(request, 'django_dump_die/sample.html', {})
    sample_template_response.render()

    # Call dump on generated variables.
    dump('')
    dump('QueryDict object (GET and POST are instances of this):')
    dump(sample_query_dict)

    dump('')
    dump('Request object:')
    dump(sample_request)

    dump('')
    dump('HttpResponse object:')
    dump(sample_http_response)

    dump('')
    dump('TemplateResponse object:')
    dump(sample_template_response)


def dump_edgecase_types():
    """Dump Edge Case Types"""
    # Call dump on problem children.
    dump('Displaying example of various edge-case output.')
    dump('')

    dump('')
    dump('Output of parens as string within dd parameters:')
    dump(')')
    dump('(')
    dump(')()))(')
    dump(')((()(')
    dump(sample_func_param('('))
    dump(sample_func_param(')'))

    dump('')
    dump('Function call examples (when also passing DumpDie-specific kwargs. Those should be excluded from output):')
    dump(sample_func_param(32), deepcopy=True)
    dump(sample_func_param(32), index_range=(0, 1))
    dump(sample_func_param(32), deepcopy=True, index_range=(0, 1))
    dump(sample_func_param(32, foo=12), deepcopy=True)

    dump('')
    dump('Function call examples (spanning multiple lines, in code):')
    dump('')
    dump(sample_func_param(
        'test_param',
        some_kwarg=False,
        extra_kwarg_1='extra_kwarg',
        extra_kwarg_2=2,
        extra_kwarg_3=3,
    ))
    dump(sample_func_param(
        'test_param',
        'extra_arg_1',
        2,
        True,
        extra_kwarg_1='extra_kwarg',
        extra_kwarg_2=2,
        extra_kwarg_3=3,
    ))

# endregion DD Display Functions



"""Constants for dump die"""

# System Imports.
import datetime
import types
from decimal import Decimal
from pathlib import PosixPath, PurePath, PurePosixPath, PureWindowsPath

# Third-Party Imports.
from django.conf import settings
from django.forms.boundfield import BoundField
from django.utils import timezone

# Imports that may not be accessible, depending on local python environment setup.
try:
    import pytz
    PYTZ_PRESENT = True
except ImportError:
    PYTZ_PRESENT = False
try:
    from zoneinfo import ZoneInfo
    ZONEINFO_PRESENT = True
except ImportError:
    ZONEINFO_PRESENT = False


# Simple types that do not need to be recursively inspected.
SIMPLE_TYPES = [
    bool,
    BoundField,
    bytes,
    Decimal,
    float,
    int,
    types.ModuleType,
    str,
]


# Intermediate types, that need some level of recursion and some level of "simple type" handling.
INTERMEDIATE_TYPES = [
    bytearray,
    complex,

    datetime.datetime,
    datetime.date,
    datetime.time,
    datetime.timedelta,
    timezone.timezone,

    PosixPath,
    PurePath,
    PurePosixPath,
    PureWindowsPath,
]

if ZONEINFO_PRESENT:
    INTERMEDIATE_TYPES.append(ZoneInfo)


# List of additional simple types defined as strings that do not need to be recursively inspected.
ADDITIONAL_SIMPLE_TYPES = getattr(settings, 'DJANGO_DD_ADDITIONAL_SIMPLE_TYPES', [])
# List of additional intermediate types defined as strings that do not need to be recursively inspected.
ADDITIONAL_INTERMEDIATE_TYPES = getattr(settings, 'DJANGO_DD_ADDITIONAL_INTERMEDIATE_TYPES', [])
# Max recursion depth to go while processing the dumped variable.
# Deep recursion will cause DD to take forever on complex structures.
MAX_RECURSION_DEPTH = getattr(settings, 'DJANGO_DD_MAX_RECURSION_DEPTH', 5)
# Max number of entries in an iterable to process further with recursion.
# After reaching an entry beyond this length, it will just print the unique
# instead of recursing into the entry to find further details.
# EX: if set to 20, a list of 30 will recursively inspect and print out 20
# items and then simply print the unique for the last 10.
MAX_ITERABLE_LENGTH = getattr(settings, 'DJANGO_DD_MAX_ITERABLE_LENGTH', 20)
# Whether each dump should include the filename and linenumber of the dump call.
INCLUDE_FILENAME_LINENUMBER = getattr(settings, 'DJANGO_DD_INCLUDE_FILENAME_LINENUMBER', False)
# Whether attributes should be included in the output.
INCLUDE_ATTRIBUTES = getattr(settings, 'DJANGO_DD_INCLUDE_ATTRIBUTES', True)
# Whether functions should be included in the output.
INCLUDE_FUNCTIONS = getattr(settings, 'DJANGO_DD_INCLUDE_FUNCTIONS', False)
# Whether function doc output should try to fit on one line, or output with original newlines.
MULTILINE_FUNCTION_DOCS = getattr(settings, 'DJANGO_DD_MULTILINE_FUNCTION_DOCS', False)
# Whether or not to colorize the name of the dumped object
COLORIZE_DUMPED_OBJECT_NAME = getattr(settings, 'DJANGO_DD_COLORIZE_DUMPED_OBJECT_NAME', True)
# Whether objects overarching item content (Attribute, Function) should start expanded for viewing.
CONTENT_STARTS_EXPANDED = getattr(settings, 'DJANGO_DD_CONTENT_STARTS_EXPANDED', False)
# Whether the attributes for an object should start expanded for viewing.
ATTRIBUTES_START_EXPANDED = getattr(settings, 'DJANGO_DD_ATTRIBUTES_START_EXPANDED', True)
# Whether the functions for an object should start expanded for viewing.
FUNCTIONS_START_EXPANDED = getattr(settings, 'DJANGO_DD_FUNCTIONS_START_EXPANDED', False)
# Whether the output should include private attributes and functions.
INCLUDE_PRIVATE_METHODS = getattr(settings, 'DJANGO_DD_INCLUDE_PRIVATE_MEMBERS', False)
# Whether the output should include magic methods.
INCLUDE_MAGIC_METHODS = getattr(settings, 'DJANGO_DD_INCLUDE_MAGIC_METHODS', False)



"""Utils for dump die"""

# System Imports.
import copy
import inspect
import io
import linecache
import re
import tokenize
from collections.abc import Sequence
from decimal import Decimal
from enum import EnumMeta
from tokenize import (
    generate_tokens,
    ENDMARKER,
    NAME,
    NEWLINE,
    NUMBER,
    OP,
    STRING,
)

# Third-Party Imports.
from django.core.exceptions import ObjectDoesNotExist
from django.http import QueryDict

# Internal Imports.
from django_dump_die.constants import COLORIZE_DUMPED_OBJECT_NAME, INCLUDE_FILENAME_LINENUMBER, PYTZ_PRESENT

# Imports that may not be accessible, depending on local python environment setup.
if PYTZ_PRESENT:
    import pytz


class Enum:
    """Enum faker class so an entire Enum can be dumped correctly."""
    def __init__(self, *args, **kwargs):
        """Set attrs on this class for each kwarg passed in."""
        for key, val in kwargs.items():
            setattr(self, key, val)


def get_dumped_object_info(obj, index_range=None, deepcopy=False):
    """Get all the required information to be able to dump an object"""
    # Get object filename, linenumber, and name
    filename, linenumber, obj_name = get_dumped_object_name_and_location(obj)

    # Process the object name to add coloring and/or put in correct format for the template.
    obj_name = process_object_name(obj_name)

    # Handle if function.
    function_doc = None
    if callable(obj) and not inspect.isclass(obj):
        function_doc = inspect.getdoc(obj)

    # Sanitize and validate provided index values.
    start_index, end_index = sanitize_index_range(index_range)

    # Handle if Enum by converting to a standard class.
    # NOTE: Dumping an Enum without conversion results in a blank string being dumped.
    if isinstance(obj, EnumMeta):
        obj = Enum(**obj.__members__)

    # Handle if deepcopy set.
    original_obj = None
    try:
        if deepcopy:
            original_obj = obj
            obj = copy.deepcopy(obj)
    except TypeError as type_error:
        original_obj = None
        raise TypeError(f"Object contains type that can't be deep copied. - {type_error}") from None

    return (filename, linenumber, obj_name, obj, function_doc, start_index, end_index, original_obj)


def get_dumped_object_name_and_location(object_needing_name):
    """Look at the stack frame to figure out what the dumped var is"""

    # Get the frame where the dump or dd occurred.
    base_frame = inspect.currentframe().f_back.f_back.f_back
    frame_info = inspect.getframeinfo(base_frame)

    # Default to not showing filename and lineno via setting to None
    linenumber = None
    filename = None
    # If we are to show the filename and linenumber
    if INCLUDE_FILENAME_LINENUMBER:
        # Get line number
        linenumber = frame_info.lineno
        # Get filename
        filename = frame_info.filename

    # Parse line that called, for full "name" to output.
    # Accounts for things like functions that span multiple lines.
    code_context = get_fully_qualified_dumped_line(base_frame, frame_info)

    # TODO: This work of using RegEx to get down to the dumped object name can probably be improved.
    # Establish a couple of regular expressions to fetch out what was passed to dump or dd.
    dump_pattern = r".*dump\((.*)[\)]"
    dd_pattern = r".*dd\((.*)[\)]"
    options_pattern = r"(.*)(?=,.*=.*)"
    # Find the results for both dump and dd.
    dumped_text_matches = re.findall(dump_pattern, code_context)
    dd_text_matches = re.findall(dd_pattern, code_context)
    # Determine if dumped or dd'd and put result in dumped_text
    dumped_text = 'Unknown_Object_Name'
    if dumped_text_matches:
        dumped_text = dumped_text_matches[0]
    elif dd_text_matches:
        dumped_text = dd_text_matches[0]

    while 'deepcopy' in dumped_text or 'index_range' in dumped_text:
        dumped_text_matches = re.findall(options_pattern, dumped_text)
        if dumped_text_matches:
            dumped_text = dumped_text_matches[0]
        else:
            break

    # If function get the callable name for each function name.
    if callable(object_needing_name) and not inspect.isclass(object_needing_name):
        dumped_text = get_callable_name(dumped_text, object_needing_name)

    return filename, linenumber, dumped_text


def get_fully_qualified_dumped_line(base_frame, frame_info):
    """Loop through frame info until we get full object name.

    Required for definitions of things that span multiple lines, within a given dump/dd statement.
    """

    line_num = base_frame.f_lineno - 1
    code_context = ''
    counter = None
    total_loops = 0
    # Loop through lines until counter goes back to 0, to account for newlines.
    # Counter increments for every proper ( token found, and decrements for every proper ) token found.
    # Additional checks are present so that the page does not hang forever, in event of input we did not account for.
    while counter is None or (0 < counter < 100 and total_loops < 1000):
        total_loops += 1

        # Initialize counter for first loop.
        if counter is None:
            counter = 0

        # Check line for parsing actual name info.
        line_num += 1
        line = linecache.getline(frame_info.filename, line_num, base_frame.f_globals)
        code_context += line

        # Get the tokens out of the string.
        # NOTE: generate_tokens returns an iterator and not a data structure.
        tokens = generate_tokens(io.StringIO(line).readline)

        # Iterate through all tokens from line, and adjust counter if paren tokens are found.
        try:
            for token in tokens:
                if token.exact_type == 7:
                    counter += 1
                elif token.exact_type == 8:
                    counter -= 1

        except tokenize.TokenError:
            # Seems to only come up for end of line. We can safely ignore, probably?
            pass

    # Strip out extra whitespace, if present.
    code_context = re.sub(r'\s+', ' ', code_context)

    # Return final result.
    return code_context


def process_object_name(object_name):
    """Process the object name into a dictionary of parts that can be colorized in output"""

    # List to return for the name
    name = []

    # If the dumped object name should be colorized
    if COLORIZE_DUMPED_OBJECT_NAME:

        # List of function names
        functions = []

        # List of params
        params = []

        # Get the tokens out of the string
        # NOTE: generate_tokens returns an iterator and not a data structure.
        tokens = generate_tokens(io.StringIO(object_name).readline)

        # Stores all tokens extracted from the iterator.
        all_tokens = []
        # Stores the previous token as we iterate.
        previous_token = ''
        # Whether we are inside a param list. Default to False
        in_param_list = False

        # Convert the tokens from the iterator to a list of tokens.
        # Also, store the tokens that are function names or params for later color processing.
        for token in tokens:
            # Store token for later processing.
            all_tokens.append(token)
            # Convert for readability
            token_number = token[0]
            token_value = token[1]

            # If token is a opening paren
            if token_number == OP and token.exact_type == 7:
                # Append previous token to the function list
                functions.append(previous_token)
                # Flip flag to indicate that we are now processing the parameter list.
                in_param_list = True

            # If we are inside the param list and not at the ending paren
            if in_param_list and token.exact_type != 8:
                # Append the token to the list of params
                params.append(token_value)
            else:
                in_param_list = False

            # Update the previous token to the current one.
            previous_token = token_value

        # Iterate over tokens classifying them with a CSS class.
        for token_number, token_value, _, _, _ in all_tokens:

            # Ignore newlines and endmarkers
            if token_number == NEWLINE or token_number == ENDMARKER:
                continue

            if token_number == NAME and token_value == 'Unknown_Object_Name':
                # Use empty color as object name can't be determined
                css_class = 'empty'
            elif token_number == NAME and is_const(token_value):
                # Use constant color.
                css_class = 'constant'
            elif token_number == STRING:
                # Use string color.
                css_class = 'string'
            elif token_number == NUMBER:
                # Use number color.
                css_class = 'number'
            elif token_number == OP and token_value in '(){}[]':
                # Use braces color.
                css_class = 'braces'
            elif token_value in functions:
                # Use function color
                css_class = 'function'
            elif token_value in params:
                # User param color
                css_class = 'params'
            else:
                # Use default color.
                css_class = 'dumped_name'

            # Append the info on to the name list
            name.append({
                'css_class': css_class,
                'value': token_value,
            })

    else:
        # Convert object name to the format expected by the template.
        name.append({
            'css_class': '',
            'value': object_name,
        })

    return name


def sanitize_index_range(index_range):
    """
    Validates and sanitizes passed index values.
    """
    # Set index defaults.
    start_index = None
    end_index = None

    # First check if index_range is an iterable.
    if isinstance(index_range, Sequence):
        # Handle for length 2 or more (we ignore values past second index).
        if len(index_range) > 1:
            # Assume is index range. Split values onto indexes.
            start_index = index_range[0]
            end_index = index_range[1]

        # Handle for length exactly 1.
        elif len(index_range) == 1:
            # Assume user typo'd somehow and meant literal start index.
            start_index = index_range[0]

    else:
        # Non-iterable. Assume user passed single start-index value.
        start_index = index_range

    # Sanitize if start_index set.
    if start_index:
        try:
            start_index = int(start_index)
        except TypeError:
            start_index = None

    # Sanitize if end_index set.
    if end_index:
        try:
            end_index = int(end_index)
        except Exception:
            end_index = None

    return start_index, end_index


# region Object Property Functions


def generate_unique_from_obj(obj):
    """Generate a unique identifier for the object passed in."""

    # Create unique via hash and fallback to id on exception.
    try:
        unique = hash(obj)
    except Exception:
        unique = id(obj)
    # Append the class name to the unique to really make unique.
    unique = f'{get_class_name(obj)}_{unique}'

    return unique


def get_members(obj):
    """Attempts to get object members. Falls back to an empty list."""

    # Get initial member set or empty list.
    members = inspect.getmembers(obj)

    # Add type specific members that will not be included from the use of the inspect.getmembers function.
    if is_dict(obj):
        if isinstance(obj, QueryDict):
            # Django QueryDict members. Has handling for multiple unique values referencing the same key.
            # https://docs.djangoproject.com/en/dev/ref/request-response/#querydict-objects
            obj_copy = copy.deepcopy(obj)
            for key in obj.keys():
                item = obj_copy.pop(key)
                members.append((key, item))
        else:
            # Standard dictionary members.
            members.extend(obj.items())

    elif is_iterable(obj):
        # Lists, sets, etc.
        if _is_indexable(obj):
            # Use indexes as left half.
            members.extend(list(enumerate(obj)))
        elif is_set(obj):
            # Use None as left half. Most likely a set.
            members.extend([(None, x) for x in obj])

    return members


def get_class_name(obj):
    """Get class name of an object."""
    name = None
    try:
        name = obj.__class__.__name__
    except Exception:
        pass
    return name


def get_callable_name(obj_name, obj):
    """Get callable name of an object"""

    # Get the method signature and fall back to simply appending
    # parentheses to the method name on exception.
    try:
        obj_name += safe_str(inspect.signature(obj))
    except Exception:
        obj_name += '()'
    return obj_name


def get_callable_params(obj):
    """Get callable params of an object"""

    # Get the method signature and fall back to simply returning
    # parentheses on exception.
    try:
        signature = safe_str(inspect.signature(obj))
    except Exception:
        signature = '()'

    signature_minus_parentheses = signature[1:-1]

    return signature_minus_parentheses


def get_obj_type(obj):
    """Determines the string representation of object's type."""

    # Get default type value.
    obj_type = type(obj).__name__

    # Special handling for certain types.
    if obj_type == 'NoneType':
        obj_type = 'null'
    elif PYTZ_PRESENT and isinstance(obj, pytz.BaseTzInfo):
        obj_type = 'pytz_timezone'

    return obj_type


def safe_repr(obj):
    """Call repr() and ignore ObjectDoesNotExist."""
    str_obj = ''
    try:
        str_obj = repr(obj)
    except ObjectDoesNotExist:
        # NOTE: A list of deleted db objects will cause repr(list) to fail.
        # So, we detect this and print out the __class__ of the contents of
        # the list.
        str_obj = f'<{obj.__class__} DELETED>'

    return str_obj


def safe_str(obj):
    """Call str() and ignore TypeErrors if str() doesn't return a string."""
    str_obj = ''
    try:
        str_obj = str(obj)
    except (TypeError, ObjectDoesNotExist):
        str_obj = safe_repr(obj)

    return str_obj


def is_iterable(obj):
    """Return True if object can be iterated."""
    try:
        iter(obj) and len(obj)
    except NotImplementedError:
        return False
    except TypeError:
        return False
    except ValueError:
        return False
    return True


def is_query(obj):
    """Return True if object is most likely a query."""
    if obj is not None:
        return _in_dir(obj, 'as_manager') and _in_dir(obj, 'all') and _in_dir(obj, 'filter')


def is_dict(obj):
    """Return True if object is most likely a dict."""
    if obj is not None:
        return _in_dir(obj, 'items') and _in_dir(obj, 'keys') and _in_dir(obj, 'values')


def is_set(obj):
    """Return True if object is most likely a dict."""
    return isinstance(obj, set) or isinstance(obj, frozenset)


def is_const(obj):
    """Return True if object is most likely a constant."""
    if obj is not None:
        return isinstance(obj, str) and obj[0].isalpha() and obj.upper() == obj


def is_key(obj):
    """Return True if object is most likely a key."""
    if obj is not None and isinstance(obj, str):
        return "'" in obj


def is_number(obj):
    """Return True if object is most likely a number."""
    if obj is not None:
        return (
            isinstance(obj, (int, float, Decimal))
            or (getattr(obj, 'isnumeric', None) and obj.isnumeric())
        )


def is_private(obj):
    """Return True if object is private."""
    if obj is not None:
        return isinstance(obj, str) and obj.startswith('_')


def is_magic(obj):
    """Return True if object is private."""
    if obj is not None:
        return isinstance(obj, str) and obj.startswith('__') and obj.endswith('__')


def _in_dir(obj, attr):
    """Simpler hasattr() function without side effects."""
    return attr in dir(obj)


def _is_indexable(obj):
    """Return True if object can be indexed."""
    if isinstance(obj, Sequence):
        return True
    else:
        return False

# endregion Object Property Functions



"""
Tests for "request/response cycle" type DD output.
"""

# Third-Party Imports.
from django.test import override_settings
from django_expanded_test_cases import IntegrationTestCase


@override_settings(DEBUG=True)
class DumpDieDjangoRequestResponseCycleTestCase(IntegrationTestCase):
    """Verify handling of dumped "Django request-response cycle" types."""
    url = 'django_dump_die:django-request-response-cycle-example'

    def test_toolbar_display(self):
        """Verify page properly displays toolbar."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                # Check toolbar header.
                '<div class="dump-toolbar">',
                '<div><h1>Django DumpDie</h1></div>',

                # Check existence of buttons.
                '<p id="expand-all" class="button">Expand All</p>',
                '<p id="expand-1st-lvl" class="button">Expand First Level</p>',
                '<p id="expand-2nd-lvl" class="button">Expand Second Level</p>',
                '<p id="collapse-all" class="button">Collapse All</p>',
                '<p id="collapse-1st-lvl" class="button">Collapse First Level</p>',
                '<p id="collapse-2nd-lvl" class="button">Collapse Second Level</p>',

                '<div class="static-padding"></div>',
            ],
            content_starts_after='<body>',
            content_ends_before='<div class="dump-wrapper">',
        )

    def test_page_descriptor_display(self):
        """Verify initial page descriptor output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                # Check page descriptor.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">'Displaying Django request-response-cycle example output.'</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'Displaying Django request-response-cycle example output.'</code>
                </div>
                """,
                '<hr>',
                # Check visual-padding lines.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">''</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </div>
                """,
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">''</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='<div class="static-padding"></div>',
            content_ends_before='QueryDict object',
        )

    # TODO/NOTES:
    #   Due to the number of "unique" values being generated in a Request object, it seems to error out with these
    #   tests. Either we get a StopIteration error, due to an insufficient mock range, or we get a RecursionError
    #   due to too many calls when we raise this mock value.
    #
    #   At the moment, I'm unsure of if there's a way around this to physically test this page logic.
    #
    #   Note: The below QueryDict test technically passes (at least when this was written, as of early Aug 2022) IF you
    #   edit the view itself to stop further dd output after displaying the QueryDict. But as soon as you include the
    #   Request object, then the above described errors occur.
    #
    #   And at least for now, I'd rather have all output types displayed for manual user-examination when physically
    #   loading this page in a browser, rather than comment out the "Request object dump", just for the sake of
    #   UnitTests.
    #
    #
    # @patch('django_dump_die.templatetags.dump_die._generate_unique')
    # def test_querydict_display(self, mocked_unique_generation):
    #     """Verify dumping a "QueryDict" type has expected output."""
    #     # Override default "unique" generation logic, for reproduce-able tests.
    #     # This generates enough uniques to guarantee mock does not raise errors.
    #     # Unlike above tests, we need to give each object custom uniques, to ensure child-objects display.
    #     side_effects = []
    #     for index in range(5000):
    #         side_effects += [
    #             (f'data_900{index}', ''),
    #         ]
    #     mocked_unique_generation.side_effect = side_effects
    #
    #     # Test object display.
    #     self.assertGetResponse(
    #         self.url,
    #         expected_title='DD',
    #         expected_header='Django DumpDie',
    #         expected_content=[
    #             '<hr>',
    #
    #             # Object descriptor.
    #             """
    #             <div class="dump-wrapper">
    #                 <span class="dumped_object" title="Dumped Object">
    #                     <span class="string">'QueryDict object (GET and POST are instances of this):'</span>
    #                 </span>:
    #                 <span class="type" title="str">str</span>
    #                 <code class="string">'QueryDict object (GET and POST are instances of this):'</code>
    #             </div>
    #             """,
    #
    #             '<hr>',
    #             # Object opening tags.
    #             """
    #             <div class="dump-wrapper">
    #                 <span class="dumped_object" title="Dumped Object">
    #                     <span class="dumped_name">sample_query_dict</span>
    #                 </span>:
    #                 <span class="type" title="QueryDict">QueryDict:0</span>
    #                 <span class="braces">{</span>
    #                 <a
    #                     class="arrow-toggle collapsed"
    #                     title="[Ctrl+click] Expand all children"
    #                     data-toggle="collapse"
    #                     data-target=".data_9008"
    #                     data-dd-type="type"
    #                     data-object-depth="1"
    #                     aria-label="Close"
    #                     aria-expanded="false"
    #                 >
    #                     <span class="unique" data-highlight-unique="data_9008">data_9008</span>
    #                     <span id="arrow-data_9008" class="arrow arrow-data_9008">▶</span>
    #                 </a>
    #                 <div class="dd-wrapper collapse data_9008 " data-unique="data_9008">
    #                 <ul class="attribute-list">
    #                 <a
    #                     class="arrow-toggle show always-show"
    #                     title="[Ctrl+click] Expand all children"
    #                     data-target=".data_9008-attributes"
    #                     data-dd-type="attr"
    #                     aria-label="Open/Close"
    #                     aria-expanded=""
    #                 >
    #                     <span class="section_name">Attributes</span>
    #                         <span id="arrow-data_9008-attributes" class="arrow">
    #                     </span>
    #                 </a>
    #                 <div
    #                     class="li-wrapper collapse data_9008-attributes show"
    #                     data-unique-attributes="data_9008-attributes"
    #                 >
    #             """,
    #
    #             # Object child elements.
    #             """
    #             <li>
    #                 <span class="key" title="Key">'one_val'</span>:
    #                 <span class="type" title="list">list:1</span>
    #                 <span class="braces">[</span>
    #                 <a
    #                     class="arrow-toggle collapsed"
    #                     title="[Ctrl+click] Expand all children"
    #                     data-toggle="collapse"
    #                     data-target=".data_90010"
    #                     data-dd-type="type"
    #                     data-object-depth="2"
    #                     aria-label="Close"
    #                     aria-expanded="false"
    #                 >
    #                     <span class="unique" data-highlight-unique="data_90010">data_90010</span>
    #                     <span id="arrow-data_90010" class="arrow arrow-data_90010">▶</span>
    #                 </a>
    #                 <div class="dd-wrapper collapse data_90010 " data-unique="data_90010">
    #                     <ul class="attribute-list">
    #                         <a
    #                             class="arrow-toggle show always-show"
    #                             title="[Ctrl+click] Expand all children"
    #                             data-target=".data_90010-attributes"
    #                             data-dd-type="attr"
    #                             aria-label="Open/Close"
    #                             aria-expanded=""
    #                         >
    #                             <span class="section_name">Attributes</span>
    #                             <span id="arrow-data_90010-attributes" class="arrow"></span>
    #                         </a>
    #                         <div
    #                             class="li-wrapper collapse data_90010-attributes show"
    #                             data-unique-attributes="data_90010-attributes"
    #                         >
    #                             <li>
    #                                 <span class="index" title="Index">0</span>:
    #                                 <span class="type" title="str">str</span> <code class="string">'test'</code>
    #                             </li>
    #                         </div>
    #                     </ul>
    #                     <ul class="attribute-list"></ul>
    #                 </div>
    #                 <span class="braces">]</span>
    #             </li>
    #             """,
    #             """
    #             <li>
    #                 <span class="key" title="Key">'two_vals'</span>:
    #                 <span class="type" title="list">list:2</span>
    #                 <span class="braces">[</span>
    #                 <a
    #                     class="arrow-toggle collapsed"
    #                     title="[Ctrl+click] Expand all children"
    #                     data-toggle="collapse"
    #                     data-target=".data_90012"
    #                     data-dd-type="type"
    #                     data-object-depth="2"
    #                     aria-label="Close"
    #                     aria-expanded="false"
    #                 >
    #                     <span class="unique" data-highlight-unique="data_90012">data_90012</span>
    #                     <span id="arrow-data_90012" class="arrow arrow-data_90012">▶</span>
    #                 </a>
    #                 <div class="dd-wrapper collapse data_90012 " data-unique="data_90012">
    #                     <ul class="attribute-list">
    #                         <a
    #                             class="arrow-toggle show always-show"
    #                             title="[Ctrl+click] Expand all children"
    #                             data-target=".data_90012-attributes"
    #                             data-dd-type="attr"
    #                             aria-label="Open/Close"
    #                             aria-expanded=""
    #                         >
    #                             <span class="section_name">Attributes</span>
    #                             <span id="arrow-data_90012-attributes" class="arrow"></span>
    #                         </a>
    #                         <div
    #                             class="li-wrapper collapse data_90012-attributes show"
    #                             data-unique-attributes="data_90012-attributes"
    #                         >
    #                             <li>
    #                                 <span class="index" title="Index">0</span>:
    #                                 <span class="type" title="str">str</span> <code class="string">'one'</code>
    #                             </li>
    #                             <li>
    #                                 <span class="index" title="Index">1</span>:
    #                                 <span class="type" title="str">str</span> <code class="string">'2'</code>
    #                             </li>
    #                         </div>
    #                     </ul>
    #                     <ul class="attribute-list"></ul>
    #                 </div>
    #                 <span class="braces">]</span>
    #             </li>
    #             """,
    #             """
    #             <li>
    #                 <span class="key" title="Key">'example_field_list'</span>:
    #                 <span class="type" title="list">list:3</span>
    #                 <span class="braces">[</span>
    #                 <a
    #                     class="arrow-toggle collapsed"
    #                     title="[Ctrl+click] Expand all children"
    #                     data-toggle="collapse"
    #                     data-target=".data_90015"
    #                     data-dd-type="type"
    #                     data-object-depth="2"
    #                     aria-label="Close"
    #                     aria-expanded="false"
    #                 >
    #                     <span class="unique" data-highlight-unique="data_90015">data_90015</span>
    #                     <span id="arrow-data_90015" class="arrow arrow-data_90015">▶</span>
    #                 </a>
    #                 <div class="dd-wrapper collapse data_90015 " data-unique="data_90015">
    #                     <ul class="attribute-list">
    #                         <a
    #                             class="arrow-toggle show always-show"
    #                             title="[Ctrl+click] Expand all children"
    #                             data-target=".data_90015-attributes"
    #                             data-dd-type="attr"
    #                             aria-label="Open/Close"
    #                             aria-expanded=""
    #                         >
    #                             <span class="section_name">Attributes</span>
    #                             <span id="arrow-data_90015-attributes" class="arrow"></span>
    #                         </a>
    #                         <div
    #                             class="li-wrapper collapse data_90015-attributes show"
    #                             data-unique-attributes="data_90015-attributes"
    #                         >
    #                             <li>
    #                                 <span class="index" title="Index">0</span>:
    #                                 <span class="type" title="str">str</span> <code class="string">'username'</code>
    #                             </li>
    #                             <li>
    #                                 <span class="index" title="Index">1</span>:
    #                                 <span class="type" title="str">str</span> <code class="string">'first_name'</code>
    #                             </li>
    #                             <li>
    #                                 <span class="index" title="Index">2</span>:
    #                                 <span class="type" title="str">str</span> <code class="string">'last_name'</code>
    #                             </li>
    #                         </div>
    #                     </ul>
    #                     <ul class="attribute-list"></ul>
    #                 </div>
    #                 <span class="braces">]</span>
    #             </li>
    #             """,
    #             """
    #             <li>
    #                 <span class="key" title="Key">'example_types'</span>:
    #                 <span class="type" title="list">list:4</span>
    #                 <span class="braces">[</span>
    #                 <a
    #                     class="arrow-toggle collapsed"
    #                     title="[Ctrl+click] Expand all children"
    #                     data-toggle="collapse"
    #                     data-target=".data_90019"
    #                     data-dd-type="type"
    #                     data-object-depth="2"
    #                     aria-label="Close"
    #                     aria-expanded="false"
    #                 >
    #                     <span class="unique" data-highlight-unique="data_90019">data_90019</span>
    #                     <span id="arrow-data_90019" class="arrow arrow-data_90019">▶</span>
    #                 </a>
    #                 <div class="dd-wrapper collapse data_90019 " data-unique="data_90019">
    #                     <ul class="attribute-list">
    #                         <a
    #                             class="arrow-toggle show always-show"
    #                             title="[Ctrl+click] Expand all children"
    #                             data-target=".data_90019-attributes"
    #                             data-dd-type="attr"
    #                             aria-label="Open/Close"
    #                             aria-expanded=""
    #                         >
    #                             <span class="section_name">Attributes</span>
    #                             <span id="arrow-data_90019-attributes" class="arrow"></span>
    #                         </a>
    #                         <div
    #                             class="li-wrapper collapse data_90019-attributes show"
    #                             data-unique-attributes="data_90019-attributes"
    #                         >
    #                             <li>
    #                                 <span class="index" title="Index">0</span>:
    #                                 <span class="type" title="null">null</span> <code class="none">None</code>
    #                             </li>
    #                             <li>
    #                                 <span class="index" title="Index">1</span>:
    #                                 <span class="type" title="bool">bool</span> <code class="bool">True</code>
    #                             </li>
    #                             <li>
    #                                 <span class="index" title="Index">2</span>:
    #                                 <span class="type" title="int">int</span> <code class="number">5</code>
    #                             </li>
    #                             <li>
    #                                 <span class="index" title="Index">3</span>:
    #                                 <span class="type" title="float">float</span> <code class="number">3.0</code>
    #                             </li>
    #                         </div>
    #                     </ul>
    #                     <ul class="attribute-list"></ul>
    #                 </div>
    #                 <span class="braces">]</span>
    #             </li>
    #             """,
    #
    #             # Object closing tags.
    #             """
    #                                 </div>
    #                             </ul>
    #                         <ul class="attribute-list"></ul>
    #                     </div>
    #                 <span class="braces">}</span>
    #             </div>
    #             """,
    #
    #             '<hr>',
    #         ],
    #         content_starts_after='QueryDict object',
    #         content_ends_before='Request object',
    #     )

    # @patch('django_dump_die.templatetags.dump_die._generate_unique')
    # def test_request_display(self, mocked_unique_generation):
    #     """Verify dumping a "Request" type has expected output."""
    #     # Override default "unique" generation logic, for reproduce-able tests.
    #     # This generates enough uniques to guarantee mock does not raise errors.
    #     # Unlike above tests, we need to give each object custom uniques, to ensure child-objects display.
    #     side_effects = []
    #     for index in range(5000):
    #         side_effects += [
    #             (f'data_900{index}', ''),
    #         ]
    #     mocked_unique_generation.side_effect = side_effects
    #
    #     # Test object display.
    #     self.assertGetResponse(
    #         self.url,
    #         expected_title='DD',
    #         expected_header='Django DumpDie',
    #         expected_content=[
    #             '<hr>',
    #
    #             # Object opening tags.
    #             """
    #             <div class="dump-wrapper">
    #             """,
    #
    #             # Object child elements.
    #             """
    #             <li>
    #             </li>
    #             """,
    #             """
    #             <li>
    #             </li>
    #             """,
    #             """
    #             <li>
    #             </li>
    #             """,
    #
    #             # Object closing tags.
    #             """
    #                                 </div>
    #                             </ul>
    #                         <ul class="attribute-list"></ul>
    #                     </div>
    #                 <span class="braces">]</span>
    #             </div>
    #             """,
    #
    #             '<hr>',
    #         ],
    #         content_starts_after='',
    #         # content_ends_before='',
    #     )

    # @patch('django_dump_die.templatetags.dump_die._generate_unique')
    # def test_http_response_display(self, mocked_unique_generation):
    #     """Verify dumping a "HttpResponse" type has expected output."""
    #     # Override default "unique" generation logic, for reproduce-able tests.
    #     # This generates enough uniques to guarantee mock does not raise errors.
    #     # Unlike above tests, we need to give each object custom uniques, to ensure child-objects display.
    #     side_effects = []
    #     for index in range(5000):
    #         side_effects += [
    #             (f'data_900{index}', ''),
    #         ]
    #     mocked_unique_generation.side_effect = side_effects
    #
    #     self.assertGetResponse(
    #         self.url,
    #         expected_title='DD',
    #         expected_header='Django DumpDie',
    #         expected_content=[
    #             '<hr>',
    #
    #             # Object opening tags.
    #             """
    #             <div class="dump-wrapper">
    #             """,
    #
    #             # Object child elements.
    #             """
    #             <li>
    #             </li>
    #             """,
    #             """
    #             <li>
    #             </li>
    #             """,
    #             """
    #             <li>
    #             </li>
    #             """,
    #
    #             # Object closing tags.
    #             """
    #                                 </div>
    #                             </ul>
    #                         <ul class="attribute-list"></ul>
    #                     </div>
    #                 <span class="braces">]</span>
    #             </div>
    #             """,
    #
    #             '<hr>',
    #         ],
    #         content_starts_after='',
    #         # content_ends_before='',
    #     )

    # @patch('django_dump_die.templatetags.dump_die._generate_unique')
    # def test_template_response_display(self, mocked_unique_generation):
    #     """Verify dumping a "TemplateResponse" type has expected output."""
    #     # Override default "unique" generation logic, for reproduce-able tests.
    #     # This generates enough uniques to guarantee mock does not raise errors.
    #     # Unlike above tests, we need to give each object custom uniques, to ensure child-objects display.
    #     side_effects = []
    #     for index in range(5000):
    #         side_effects += [
    #             (f'data_900{index}', ''),
    #         ]
    #     mocked_unique_generation.side_effect = side_effects
    #     self.assertGetResponse(
    #         self.url,
    #         expected_title='DD',
    #         expected_header='Django DumpDie',
    #         expected_content=[
    #             '<hr>',
    #
    #             # Object opening tags.
    #             """
    #             <div class="dump-wrapper">
    #             """,
    #
    #             # Object child elements.
    #             """
    #             <li>
    #             </li>
    #             """,
    #             """
    #             <li>
    #             </li>
    #             """,
    #             """
    #             <li>
    #             </li>
    #             """,
    #
    #             # Object closing tags.
    #             """
    #                                 </div>
    #                             </ul>
    #                         <ul class="attribute-list"></ul>
    #                     </div>
    #                 <span class="braces">]</span>
    #             </div>
    #             """,
    #
    #             '<hr>',
    #         ],
    #         content_starts_after='',
    #         # content_ends_before='',
    #     )



"""
Tests for "function" type DD output.
"""

# Third-Party Imports.
from django.test import override_settings
from django_expanded_test_cases import IntegrationTestCase


@override_settings(DEBUG=True)
class DumpDieViewFunctionTestCase(IntegrationTestCase):
    """Verify handling of dumped function types."""
    url = 'django_dump_die:function-example'

    def test_toolbar_display(self):
        """Verify page properly displays toolbar."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                # Check toolbar header.
                '<div class="dump-toolbar">',
                '<div><h1>Django DumpDie</h1></div>',

                # Check existence of buttons.
                '<p id="expand-all" class="button">Expand All</p>',
                '<p id="expand-1st-lvl" class="button">Expand First Level</p>',
                '<p id="expand-2nd-lvl" class="button">Expand Second Level</p>',
                '<p id="collapse-all" class="button">Collapse All</p>',
                '<p id="collapse-1st-lvl" class="button">Collapse First Level</p>',
                '<p id="collapse-2nd-lvl" class="button">Collapse Second Level</p>',

                '<div class="static-padding"></div>',
            ],
            content_starts_after='<body>',
            content_ends_before='<div class="dump-wrapper">',
        )

    def test_page_descriptor_display(self):
        """Verify initial page descriptor output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                # Check page descriptor.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">'Displaying example of function object output.'</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'Displaying example of function object output.'</code>
                </div>
                """,
                '<hr>',
                # Check visual-padding lines.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">''</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </div>
                """,
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">''</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='<div class="static-padding"></div>',
            content_ends_before='<span class="string">\'Function examples:\'</span>',
        )

    def test_basic_func_display(self):
        """Verify dumping a "basic function" type (no args) has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_function" title="Dumped Function">
                        <span class="function">sample_func</span>
                        <span class="braces">(</span>
                        <span class="braces">)</span>
                    </span>:
                    <span class="docs">Sample doc string</span>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='Function examples:',
            content_ends_before='sample_func_param',
        )

    def test_func_with_args_display(self):
        """Verify dumping a "function" type (with args/kwargs) has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_function" title="Dumped Function">
                        <span class="function">sample_func_param</span><span class="braces">(</span>
                        <span class="params">param1</span><span class="params">,</span>
                        <span class="params">*</span>
                        <span class="params">args</span>
                        <span class="params">,</span>
                        <span class="params">some_kwarg</span>
                        <span class="params">=</span>
                        <span class="params">None</span>
                        <span class="params">,</span>
                        <span class="params">**</span>
                        <span class="params">kwargs</span>
                        <span class="braces">)</span>
                    </span>:
                    <span class="docs">Sample param doc string. :param param1: Doc for param1.</span>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='sample_func',
            content_ends_before='Function call examples:',
        )

    def test_basic_func_call_display(self):
        """Verify dumping a called "basic function" type (no args) has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="function">sample_func</span>
                        <span class="braces">(</span>
                        <span class="braces">)</span>
                    </span>:
                    <span class="type" title="int">int</span>
                    <code class="number">42</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='Function call examples:',
            content_ends_before="""
                <span class="function">sample_func_param</span>
                    <span class="braces">(</span>
                    <span class="number">32</span>
                    <span class="braces">)</span>
                </span>
            """,
        )

    def test_func_call_with_one_arg_display(self):
        """Verify dumping a called "basic function" type (one arg) has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="function">sample_func_param</span>
                            <span class="braces">(</span>
                            <span class="number">32</span>
                            <span class="braces">)</span>
                        </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'MyReturnValue with param1 as: 32'</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after="""
            <span class="function">sample_func</span>
            <span class="braces">(</span>
            <span class="braces">)</span>
            """,
            content_ends_before="""
            <span class="function">sample_func_param</span>
            <span class="braces">(</span>
            <span class="string">'test_param'</span>
            <span class="params">,</span>
            <span class="params">some_kwarg</span>
            <span class="params">=</span>
            <span class="params">True</span>
            <span class="braces">)</span>
            """,
        )

    def test_func_call_with_args_kwargs_display(self):
        """Verify dumping a called function type (with args & kwargs) has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="function">sample_func_param</span>
                        <span class="braces">(</span>
                        <span class="string">'test_param'</span>
                        <span class="params">,</span>
                        <span class="params">some_kwarg</span>
                        <span class="params">=</span>
                        <span class="params">True</span>
                        <span class="braces">)</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'MyReturnValue with param1 as: test_param'</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after="""
            <span class="function">sample_func_param</span>
            <span class="braces">(</span>
            <span class="number">32</span>
            <span class="braces">)</span>
            """,
            content_ends_before="""
            <span class="function">sample_func_param</span>
            <span class="braces">(</span>
            <span class="string">'test_param'</span>
            <span class="params">,</span>
            <span class="string">'extra_arg_1'</span>
            <span class="params">,</span>
            <span class="number">2</span>
            <span class="params">,</span>
            <span class="params">True</span>
            <span class="braces">)</span>
            """,
        )

    def test_func_call_with_multiple_args_display(self):
        """Verify dumping a called function type (with multiple args) has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="function">sample_func_param</span>
                        <span class="braces">(</span>
                        <span class="string">'test_param'</span>
                        <span class="params">,</span>
                        <span class="string">'extra_arg_1'</span>
                        <span class="params">,</span>
                        <span class="number">2</span>
                        <span class="params">,</span>
                        <span class="params">True</span>
                        <span class="braces">)</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'MyReturnValue with param1 as: test_param'</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after="""
            <span class="dumped_object" title="Dumped Object">
                <span class="function">sample_func_param</span>
                <span class="braces">(</span>
                <span class="string">'test_param'</span>
                <span class="params">,</span>
                <span class="params">some_kwarg</span>
                <span class="params">=</span>
                <span class="params">True</span>
                <span class="braces">)</span>
            </span>
            """,
            content_ends_before='done',
        )



"""
Tests for "simple" type DD output.
"""

# Third-Party Imports.
from django.test import override_settings
from django_expanded_test_cases import IntegrationTestCase


@override_settings(DEBUG=True)
class DumpDieSimpleTypeTestCase(IntegrationTestCase):
    """Verify handling of dumped "simple" types."""
    url = 'django_dump_die:simple-type-example'

    def test_toolbar_display(self):
        """Verify page properly displays toolbar."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                # Check toolbar header.
                '<div class="dump-toolbar">',
                '<div><h1>Django DumpDie</h1></div>',

                # Check existence of buttons.
                '<p id="expand-all" class="button">Expand All</p>',
                '<p id="expand-1st-lvl" class="button">Expand First Level</p>',
                '<p id="expand-2nd-lvl" class="button">Expand Second Level</p>',
                '<p id="collapse-all" class="button">Collapse All</p>',
                '<p id="collapse-1st-lvl" class="button">Collapse First Level</p>',
                '<p id="collapse-2nd-lvl" class="button">Collapse Second Level</p>',

                '<div class="static-padding"></div>',
            ],
            content_starts_after='<body>',
            content_ends_before='<div class="dump-wrapper">',
        )

    def test_page_descriptor_display(self):
        """Verify initial page descriptor output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                # Check page descriptor.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">'Displaying example of "simple type" object output.'</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'Displaying example of "simple type" object output.'</code>
                </div>
                """,
                '<hr>',
                # Check visual-padding lines.
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">''</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </div>
                """,
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="string">''</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">''</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='<div class="static-padding"></div>',
            content_ends_before='<span class="constant">SAMPLE_CONST</span>',
        )

    def test_constant_display(self):
        """Verify dumping a "constant" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="constant">SAMPLE_CONST</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'Sample Constant Content'</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='<div class="static-padding"></div>',
            content_ends_before='sample_module',
        )

    def test_module_display(self):
        """Verify dumping a "module" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_module</span>
                    </span>:
                    <span class="type" title="module">module</span>
                    <code class="module"><module 'django.html'></code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='SAMPLE_CONST',
            content_ends_before='sample_bytes',
        )

    def test_bytes_display(self):
        """Verify dumping a "bytes" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_bytes</span>
                    </span>:
                    <span class="type" title="bytes">bytes</span>
                    <code class="number">b'sample bytes'</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='sample_module',
            content_ends_before='sample_int',
        )

    def test_int_display(self):
        """Verify dumping a "int" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_int</span>
                    </span>:
                    <span class="type" title="int">int</span>
                    <code class="number">42</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='sample_bytes',
            content_ends_before='sample_float',
        )

    def test_float_display(self):
        """Verify dumping a "float" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_float</span>
                    </span>:
                    <span class="type" title="float">float</span>
                    <code class="number">42.42</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='sample_int',
            content_ends_before='sample_decimal',
        )

    def test_decimal_display(self):
        """Verify dumping a "decimal" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_decimal</span>
                    </span>:
                    <span class="type" title="Decimal">Decimal</span>
                    <code class="number">42.4200000000000017053025658242404460906982421875</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='sample_float',
            content_ends_before='sample_string',
        )

    def test_string_display(self):
        """Verify dumping a "string" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_string</span>
                    </span>:
                    <span class="type" title="str">str</span>
                    <code class="string">'Sample String Content'</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='sample_decimal',
            content_ends_before='sample_none',
        )

    def test_none_display(self):
        """Verify dumping a "None" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_none</span>
                    </span>:
                    <span class="type" title="null">null</span>
                    <code class="none">None</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='sample_string',
            content_ends_before='sample_bool',
        )

    def test_bool_display(self):
        """Verify dumping a "bool" type has expected output."""
        self.assertGetResponse(
            self.url,
            expected_title='DD',
            expected_header='Django DumpDie',
            expected_content=[
                '<hr>',
                """
                <div class="dump-wrapper">
                    <span class="dumped_object" title="Dumped Object">
                        <span class="dumped_name">sample_bool</span>
                    </span>:
                    <span class="type" title="bool">bool</span>
                    <code class="bool">True</code>
                </div>
                """,
                '<hr>',
            ],
            content_starts_after='sample_none',
            content_ends_before='done',
        )

    # # TODO: Finish this test. May require more work with django to test.
    # # def test_dd_for_bound_field(self):
    # #     response = self.client.get('/simple?type=bound_field')
    # #     expected = '<span class="dumped_object" title="Dumped Object"><span class="dumped_name">test_bound</span></span>: <code class="bound">?????</code>'
    # #     self.assertContains(response, expected, html=True)



"""
Middleware for DumpAndDie.
"""

# System Imports.
import logging

# Third-Party Imports.
from django.conf import settings

# Internal Imports.
from .views import dd_view
from django_dump_die.utils import get_dumped_object_info


logger = logging.getLogger('django_dump_die')
dump_objects = []


class DumpAndDie(Exception):
    """
    DumpAndDie Exception.
    Triggers the middleware exception logic which outputs the alternate dd debug view.
    """
    def __init__(self, obj):
        super().__init__(obj)
        self.object = obj


def dd(obj, index_range=None, deepcopy=False):
    """
    Immediately return debug template with info about objects.
    Includes any objects passed in through dump().

    Does nothing if DEBUG != True.
    """

    if settings.DEBUG:

        # Get the object info
        object_info = get_dumped_object_info(obj, index_range, deepcopy)

        # Run dd core logic.
        raise DumpAndDie(object_info)


def dump(obj, index_range=None, deepcopy=False):
    """
    Show debug template whenever response finishes.
    dd() will also include objects from dump().

    Does nothing if DEBUG != True

    NOTE: Not thread safe, this will collect objects server wide,
    dumped objects can come from multiple requests.
    """

    if settings.DEBUG:

        # Get the object info
        object_info = get_dumped_object_info(obj, index_range, deepcopy)

        # Run dd core logic.
        dump_objects.append(object_info)


class DumpAndDieMiddleware:
    """
    DumpAndDie Middleware.

    Allows access to php/laravel-like function dd().
    """
    def __init__(self, get_response):
        """
        Add our dd() and dump() commands to be universally accessible.
        """
        self.get_response = get_response

        # Add global dd() function.
        __builtins__['dd'] = dd
        # Add global dump() function.
        __builtins__['dump'] = dump

    def __call__(self, request):
        """
        Return standard response if nothing dumped.
        Otherwise return dump view.
        """
        # Get the response
        response = self.get_response(request)

        # If there are no items in the dump_objects list or there is no exception raised
        if not dump_objects or getattr(request, 'has_exception', False):
            return response
        else:
            # Create a copy of the list, and clear it.
            objects = dump_objects[:]
            dump_objects.clear()

            # Return the dd view to dump the items in the dump_objects list.
            return dd_view(request, objects)

    def process_exception(self, request, exception):
        """
        Check if exception is of DumpAndDie type.
        If so, return Debug Response.
        If not, ignore and allow standard exception handling.
        """
        if not isinstance(exception, DumpAndDie):
            request.has_exception = True
            return None

        # Create a copy of the list, and clear it.
        objects = dump_objects[:]
        objects.append(exception.object)
        dump_objects.clear()

        # Return custom DumpAndDie output view.
        return dd_view(request, objects)



"""Various test views for the DumpDie library.

Needed because having too many mocked "unique" values on a single response will lead to either:
* StopIteration - Due to giving an insufficient mock range.
* RecursionError - Due to too many mock calls, so Python seems to think we're recursing infinitely.

The solution was to divide objects out, displaying only one object per view, which lowers the total amount of
"uniques" being rendered on a given page.
"""

# Third-Party Imports.
from django.shortcuts import render


def complex_type_example__set(request):
    """Example view, rendering only "complex type" Set object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Set object output.')
    dump('')
    dump_complex_types().dump_set()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__frozen_set(request):
    """Example view, rendering only "complex type" FrozenSet object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" FrozenSet object output.')
    dump('')
    dump_complex_types().dump_frozen_set()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__tuple(request):
    """Example view, rendering only "complex type" Tuple object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Tuple object output.')
    dump('')
    dump_complex_types().dump_tuple()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__list(request):
    """Example view, rendering only "complex type" List object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" List object output.')
    dump('')
    dump_complex_types().dump_list()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__dict(request):
    """Example view, rendering only "complex type" Dict object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Dict object output.')
    dump('')
    dump_complex_types().dump_dict()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__querydict(request):
    """Example view, rendering only "complex type" QueryDict object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" QueryDict object output.')
    dump('')
    dump_complex_types().dump_querydict()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__memory_view(request):
    """Example view, rendering only "complex type" MemoryView object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" MemoryView object output.')
    dump('')
    dump_complex_types().dump_memory_view()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__enum(request):
    """Example view, rendering only "complex type" Enum object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Enum object output.')
    dump('')
    dump_complex_types().dump_enum()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__multilevel_set(request):
    """Example view, rendering only "complex type" Multi-Level Set object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Multi-Level Set object output.')
    dump('')
    dump_complex_types().dump_multilevel_set()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__multilevel_tuple(request):
    """Example view, rendering only "complex type" Multi-Level Tuple object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Multi-Level Tuple object output.')
    dump('')
    dump_complex_types().dump_multilevel_tuple()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__multilevel_list(request):
    """Example view, rendering only "complex type" Multi-Level List object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Multi-Level List object output.')
    dump('')
    dump_complex_types().dump_multilevel_list()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__multilevel_dict(request):
    """Example view, rendering only "complex type" Multi-Level Dict object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Multi-Level Dict object output.')
    dump('')
    dump_complex_types().dump_multilevel_dict()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__list_subitem(request):
    """Example view, rendering only "complex type" List sub-item object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" List sub-item object output.')
    dump('')
    dump_complex_types().dump_list_subitem()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__tuple_subitem(request):
    """Example view, rendering only "complex type" Tuple sub-item object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Tuple sub-item object output.')
    dump('')
    dump_complex_types().dump_tuple_subitem()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__tuple_subitem_func(request):
    """Example view, rendering only "complex type" Tuple sub-item function object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Tuple sub-item function object output.')
    dump('')
    dump_complex_types().dump_tuple_subitem_function()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__dict_subitem(request):
    """Example view, rendering only "complex type" Dict sub-item object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Dict sub-item object output.')
    dump('')
    dump_complex_types().dump_dict_subitem()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example__enum_subitem(request):
    """Example view, rendering only "complex type" Enum sub-item object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" Enum sub-item object output.')
    dump('')
    dump_complex_types().dump_enum_subitem()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})



"""
DjangoDumpDie example URL Configuration.
Used exclusively to show example output.
"""

# Third-Party Imports.
from django.urls import path

# Internal Imports.
from .views import (
    index,

    simple_type_example,
    intermediate_type_example,
    complex_type_example,
    function_type_example,
    class_type_example,
    full_category_example,

    numeric_example,
    datetime_example,
    django_model_example,
    iterable_group_example,
    system_path_example,
    full_purpose_example,

    django_request_response_cycle_example,
    edge_case_example,
)


app_name = 'django_dump_die'
urlpatterns = [
    # Various example views.
    path('simple-type-example/', simple_type_example, name='simple-type-example'),
    path('intermediate-type-example/', intermediate_type_example, name='intermediate-type-example'),
    path('complex-type-example/', complex_type_example, name='complex-type-example'),

    path('function-example/', function_type_example, name='function-example'),
    path('class-example/', class_type_example, name='class-example'),

    path('full-category-example/', full_category_example, name='full-category-example'),

    path('datetime-example/', datetime_example, name='datetime-example'),
    path('django-model-example/', django_model_example, name='django-model-example'),
    path('iterable-group-example/', iterable_group_example, name='iterable-group-example'),
    path('numeric-example/', numeric_example, name='numeric-example'),
    path('system-path-example/', system_path_example, name='system-path-example'),

    path('full-purpose-example/', full_purpose_example, name='full-purpose-example'),

    path(
        'django-request-response-cycle-example/',
        django_request_response_cycle_example,
        name='django-request-response-cycle-example',
    ),
    path('edge-case-example/', edge_case_example, name='edge-case-example'),

    path('', index, name='index'),
]



"""Various example views for DumpDie library. To showcase how various outputs look in practice."""

# Third-Party Imports.
from django.shortcuts import render


def index(request):
    """Index view, to easily navigate to example views."""
    return render(request, 'django_dump_die/index.html')


def simple_type_example(request):
    """Example view, rendering only "simple type" object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_simple_types

    # Output desired dump values.
    dump('Displaying example of "simple type" object output.')
    dump('')
    dump_simple_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def intermediate_type_example(request):
    """Example view, rendering only "intermediate type" object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_intermediate_types

    # Output desired dump values.
    dump('Displaying example of "intermediate type" object output.')
    dump('')
    dump_intermediate_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def complex_type_example(request):
    """Example view, rendering only "complex type" object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_complex_types

    # Output desired dump values.
    dump('Displaying example of "complex type" object output.')
    dump('')
    dump_complex_types().dump_all_objects()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def function_type_example(request):
    """Example view, rendering only function object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_function_types

    # Output desired dump values.
    dump('Displaying example of function object output.')
    dump('')
    dump_function_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def class_type_example(request):
    """Example view, rendering only class object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_class_types

    # Output desired dump values.
    dump('Displaying example of class object output.')
    dump('')
    dump_class_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def full_category_example(request):
    """Example view, rendering all examples shown in all other views, all in one page."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import (
        dump_simple_types,
        dump_intermediate_types,
        dump_complex_types,
        dump_function_types,
        dump_class_types,
    )

    # Output desired dump values.
    dump('')
    dump_simple_types()
    dump('')
    dump('')
    dump_intermediate_types()
    dump('')
    dump('')
    dump_complex_types()
    dump('')
    dump('')
    dump_function_types()
    dump('')
    dump('')
    dump_class_types()
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def datetime_example(request):
    """Example view, rendering only "datetime" object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_datetime_types

    # Output desired dump values.
    dump('Displaying example of "datetime" object output.')
    dump('')
    dump_datetime_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def django_model_example(request):
    """Example view, rendering only Django model object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_model_types

    # Output desired dump values.
    dump('Displaying example of Django model object output.')
    dump('')
    dump_model_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def iterable_group_example(request):
    """Example view, rendering only "iterable group" object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_iterable_group_types

    # Output desired dump values.
    dump('Displaying example of "iterable group" (arrays) object output.')
    dump('')
    dump_iterable_group_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def numeric_example(request):
    """Example view, rendering only "numeric" object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_numeric_types

    # Output desired dump values.
    dump('Displaying example of "numeric type" object output.')
    dump('')
    dump_numeric_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def system_path_example(request):
    """Example view, rendering only "system path" object output."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_syspath_types

    # Output desired dump values.
    dump('Displaying example of "syspath" object output.')
    dump('')
    dump_syspath_types()
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def full_purpose_example(request):
    """Example view, rendering all examples shown in all other views, all in one page."""

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import (
        dump_numeric_types,
        dump_datetime_types,
        dump_model_types,
        dump_iterable_group_types,
        dump_syspath_types,
    )

    # Output desired dump values.
    dump('')
    dump_numeric_types()
    dump('')
    dump('')
    dump_datetime_types()
    dump('')
    dump('')
    dump_model_types()
    dump('')
    dump('')
    dump_iterable_group_types()
    dump('')
    dump('')
    dump_syspath_types()
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def django_request_response_cycle_example(request):
    """"""
    from .example_helpers import dump_django_request_response_cycle_types

    # Output desired dump values.
    dump('Displaying Django request-response-cycle example output.')
    dump('')
    dump_django_request_response_cycle_types(request)
    dump('')
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})


def edge_case_example(request):
    """Example view, rendering various edge-case output.

    These are output types that don't necessarily belong in the other example views, but have
    resulted in errors/bad output, in the past.

    This view allows easily checking them to make sure they are still handled correctly.
    """

    # Import applicable helper functions/classes.
    # Imported here so that these are only loaded on view access, and not package initialization.
    from .example_helpers import dump_edgecase_types

    # Output desired dump values.
    dump('')
    dump_edgecase_types()
    dump('')

    # Force dd to prevent further view parsing.
    dd('done')

    # Show that any calls after dd() end up ignored.
    return render(request, 'django_dump_die/sample.html', {})
